#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "adgpioalert.h"

#include <gst/gst.h>
#include <gst/video/video.h>
#include <gst/video/gstvideofilter.h>
#include <glib/gstdio.h>
#include <opencv2/opencv.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/highgui.hpp>
#include "gstadmeta.h" // include gstadmeta.h for retrieving the inference results

#include <chrono>
#include <iostream>
#include <fstream>
#include <regex>
#include <string>
#include <vector>
#define PLUGIN_NAME "adgpioalert"


using namespace cv;

#define AD_GPIO_ALERT_LOCK(sample_filter) \
  (g_rec_mutex_lock(&((AdGpioAlert *)sample_filter)->priv->mutex))

#define AD_GPIO_ALERT_UNLOCK(sample_filter) \
  (g_rec_mutex_unlock(&((AdGpioAlert *)sample_filter)->priv->mutex))

GST_DEBUG_CATEGORY_STATIC(ad_gpio_alert_debug_category);
#define GST_CAT_DEFAULT ad_gpio_alert_debug_category

std::vector<std::string> split(std::string inputString)
{
    std::regex delimiter(",");
    std::vector<std::string> list(std::sregex_token_iterator(inputString.begin(), inputString.end(), delimiter, -1), std::sregex_token_iterator());
    return list;
}

// init the dio device

enum
{
  PROP_0,
  PROP_ALERT_TYPE,
  PROP_DO_STATUS,
  PROP_VOTE_COUNT,
  N_PROPERTIES
};

struct _AdGpioAlertPrivate
{
  Mat *cv_image;
  std::string alert_type;
  std::string do_status;
  int vote_count;
  void* handle;
  GRecMutex mutex;
};

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE("sink",
                                                                   GST_PAD_SINK,
                                                                   GST_PAD_ALWAYS,
                                                                   GST_STATIC_CAPS(GST_VIDEO_CAPS_MAKE("{ BGR }")));

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE("src",
                                                                  GST_PAD_SRC,
                                                                  GST_PAD_ALWAYS,
                                                                  GST_STATIC_CAPS(GST_VIDEO_CAPS_MAKE("{ BGR }")));

#define DEBUG_INIT \
  GST_DEBUG_CATEGORY_INIT(GST_CAT_DEFAULT, PLUGIN_NAME, 0, "debug category for GPIO ALERT element");

//G_DEFINE_TYPE_WITH_CODE(AdGpioAlert, ad_gpio_alert, GST_TYPE_VIDEO_FILTER, DEBUG_INIT)
G_DEFINE_TYPE_WITH_CODE(AdGpioAlert, ad_gpio_alert, GST_TYPE_VIDEO_FILTER,
                        G_ADD_PRIVATE(AdGpioAlert)
                            DEBUG_INIT)

static void ad_gpio_alert_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec);
static void ad_gpio_alert_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec);
static void ad_gpio_alert_dispose(GObject *object);
static void ad_gpio_alert_finalize(GObject *object);
static GstFlowReturn ad_gpio_alert_transform_frame_ip(GstVideoFilter *filter, GstVideoFrame *frame);
static void ad_gpio_alert_display_background(AdGpioAlert *sample_filter, Mat &mask);
static void ad_gpio_alert_initialize_images(AdGpioAlert *sample_filter, GstVideoFrame *frame, GstMapInfo &info);
static void* initial();

// register dio device
static void* initial(void* handle)
{
    handle = AVS_GetProductHandle(0);
	if (handle)
	{
		printf("initial handle ! \n");
	}
    else
        printf("fail handle ! \n");

	return handle;
}

// ************************************************************
// Required to add this gst_buffer_get_ad_batch_meta for retrieving 
// the GstAdBatchMeta from buffer
GstAdBatchMeta* gst_buffer_get_ad_batch_meta(GstBuffer* buffer)
{
    gpointer state = NULL;
    GstMeta* meta;
    const GstMetaInfo* info = GST_AD_BATCH_META_INFO;
    
    while ((meta = gst_buffer_iterate_meta (buffer, &state))) 
    {
        if (meta->info->api == info->api) 
        {
            GstAdMeta *admeta = (GstAdMeta *) meta;
            if (admeta->type == AdBatchMeta)
                return (GstAdBatchMeta*)meta;
        }
    }
    return NULL;
}
// ************************************************************

static void
ad_gpio_alert_class_init(AdGpioAlertClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstVideoFilterClass *gstvideofilter_class;

  gobject_class = (GObjectClass *)klass;
  gstvideofilter_class = (GstVideoFilterClass *)klass;
  gstelement_class = (GstElementClass *)klass;

  GST_DEBUG_CATEGORY_INIT(GST_CAT_DEFAULT, PLUGIN_NAME, 0, PLUGIN_NAME);

  gobject_class->set_property = ad_gpio_alert_set_property;
  gobject_class->get_property = ad_gpio_alert_get_property;
  gobject_class->dispose = ad_gpio_alert_dispose;
  gobject_class->finalize = ad_gpio_alert_finalize;

  g_object_class_install_property(gobject_class, PROP_ALERT_TYPE,
                                  g_param_spec_string("alert-type","alert-type",
                                                   "The alert type name when event occurred.",
                                                   "person\0",
                                                   (GParamFlags)(G_PARAM_READWRITE)));
  
  g_object_class_install_property(gobject_class, PROP_DO_STATUS,
                                  g_param_spec_string ("DO-status", "DO-status",
                                                   "set NEON DO0-DO4 ON/OFF. e.g. parameter DO-Status=1,1,0,0, DO0-DO1 ON, DO2-DO3 OFF",
                                                   "1,0,0,0",
                                                   (GParamFlags)(G_PARAM_READWRITE)));

  g_object_class_install_property(gobject_class, PROP_VOTE_COUNT,
                                  g_param_spec_int("vote-count", "vote-count",
                                                   "The number of vote_counts you want to skip", 0, 255,
                                                   10, (GParamFlags)(G_PARAM_READWRITE)));

  gst_element_class_set_static_metadata(gstelement_class,
                                        "GPIO Alert", "GstElement",
                                        "control EOS-JNX GPIO when the alert type matched in ADLINK detection result metadata. set NEON DO0-DO4 ON/OFF. e.g. parameter DO-Status=1,1,0,0, DO0-DO1 ON, DO2-DO3 OFF",
                                        "Kevinyw Lin <kevinyw.lin@adlinktech.com>");

  gst_element_class_add_pad_template(gstelement_class,
                                     gst_static_pad_template_get(&src_factory));
  gst_element_class_add_pad_template(gstelement_class,
                                     gst_static_pad_template_get(&sink_factory));

  gstvideofilter_class->transform_frame_ip =
      GST_DEBUG_FUNCPTR(ad_gpio_alert_transform_frame_ip);

  //g_type_class_add_private(klass, sizeof(AdGpioAlertPrivate));
}

static void
ad_gpio_alert_init(AdGpioAlert *
                            sample_filter)
{
  sample_filter->priv = (AdGpioAlertPrivate *)ad_gpio_alert_get_instance_private(sample_filter);
  g_rec_mutex_init(&sample_filter->priv->mutex);
  void* init_handle;
  sample_filter->priv->handle = initial(init_handle);      
  AVS_DIOSetDOState(sample_filter->priv->handle , 0, 0);  
  AVS_DIOSetDOState(sample_filter->priv->handle , 1, 0);  
  AVS_DIOSetDOState(sample_filter->priv->handle , 2, 0);  
  AVS_DIOSetDOState(sample_filter->priv->handle , 3, 0);  
}

static void
ad_gpio_alert_set_property(GObject *object, guint property_id,
                                const GValue *value, GParamSpec *pspec)
{
  AdGpioAlert *sample_filter = AD_GPIO_ALERT(object);

  AD_GPIO_ALERT_LOCK(sample_filter);

  switch (property_id)
  {
  case PROP_ALERT_TYPE:
    sample_filter->priv->alert_type = g_value_dup_string(value);
    break;
  case PROP_DO_STATUS:
    sample_filter->priv->do_status = g_value_dup_string(value);
    break;
  case PROP_VOTE_COUNT:
    sample_filter->priv->vote_count = g_value_get_int(value);
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
    break;
  }

  AD_GPIO_ALERT_UNLOCK(sample_filter);
}

static void
ad_gpio_alert_get_property(GObject *object, guint property_id,
                                GValue *value, GParamSpec *pspec)
{
  AdGpioAlert *sample_filter = AD_GPIO_ALERT(object);

  AD_GPIO_ALERT_LOCK(sample_filter);

  switch (property_id)
  {
  case PROP_ALERT_TYPE:
    g_value_set_string(value, sample_filter->priv->alert_type.c_str());
    break;
  case PROP_DO_STATUS:
    g_value_set_string(value, sample_filter->priv->do_status.c_str());
    break;
  case PROP_VOTE_COUNT:
    g_value_set_int(value, sample_filter->priv->vote_count);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
    break;
  }

  AD_GPIO_ALERT_UNLOCK(sample_filter);
}

static void
ad_gpio_alert_dispose(GObject *object)
{
  /* In dispose(), you are supposed to free all types referenced from this
   * object which might themselves hold a reference to self. Generally,
   * the most simple solution is to unref all members on which you own a 
   * reference.
   */

  /* dispose() might be called multiple times, so we must guard against
   * calling g_object_unref() on an invalid GObject by setting the member
   * NULL; g_clear_object() does this for us.
   */
}

static void
ad_gpio_alert_finalize(GObject *object)
{
  AdGpioAlert *sample_filter = AD_GPIO_ALERT(object);
  if (sample_filter->priv->cv_image != NULL)
  {
    delete sample_filter->priv->cv_image;
  }

  if (sample_filter->priv->handle != NULL)
  {
    delete sample_filter->priv->handle;
  }

  g_rec_mutex_clear(&sample_filter->priv->mutex);
}

static GstFlowReturn
ad_gpio_alert_transform_frame_ip(GstVideoFilter *filter,
                                      GstVideoFrame *frame)
{
  AdGpioAlert *sample_filter = AD_GPIO_ALERT(filter);
  GstMapInfo info;
  Mat output_image;

  gst_buffer_map(frame->buffer, &info, GST_MAP_READ);

  ad_gpio_alert_initialize_images(sample_filter, frame, info);

  AD_GPIO_ALERT_LOCK(sample_filter);
  AD_GPIO_ALERT_UNLOCK(sample_filter);

  //get alert type
  GstAdBatchMeta *meta = gst_buffer_get_ad_batch_meta(frame->buffer);
  if (meta == NULL)
      GST_MESSAGE("Adlink metadata is not exist!");
  else
  {
      AdBatch &batch = meta->batch;
      bool frame_exist = batch.frames.size() > 0 ? true : false;

      if(frame_exist)
      {
          VideoFrameData frame_info = batch.frames[0];
          int detectionResultNumber = frame_info.detection_results.size();
          int count=0;
          int DetectCount=0;

          //std::cout << "detection result number = " << detectionResultNumber << std::endl;
          for(int i = 0 ; i < detectionResultNumber ; ++i)
          {
              count = count +1;
              // find alert-type in meta data
              if (frame_info.detection_results[i].meta.find(sample_filter->priv->alert_type) != std::string::npos){
                //std::cout << "alerttype =  " << frame_info.detection_results[i].meta << "count=" << count << "DetectCount=" << DetectCount << std::endl;
                DetectCount = DetectCount +1;
              }
              // control IO
              if ( (count >= sample_filter->priv->vote_count ) && ( DetectCount >= sample_filter->priv->vote_count/2) ){
                count=0;
                DetectCount=0;
                // control io
                std::vector<std::string> ratioVec = split(sample_filter->priv->do_status.c_str());
                AVS_DIOSetDOState(sample_filter->priv->handle , 0, std::atof(ratioVec[0].c_str()) );  
                AVS_DIOSetDOState(sample_filter->priv->handle , 1, std::atof(ratioVec[1].c_str()) );
                AVS_DIOSetDOState(sample_filter->priv->handle , 2, std::atof(ratioVec[2].c_str()) );
                AVS_DIOSetDOState(sample_filter->priv->handle , 3, std::atof(ratioVec[3].c_str()) );
                //std::cout << "do_status= " << sample_filter->priv->do_status.c_str() << std::endl;
                //std::cout << "abcd =" << std::atof(ratioVec[0].c_str()) << std::atof(ratioVec[1].c_str()) << std::atof(ratioVec[2].c_str()) << std::atof(ratioVec[3].c_str()) <<  std::endl;
              }
          }

      }
  }
  
  gst_buffer_unmap(frame->buffer, &info);
  return GST_FLOW_OK;
}

static void
ad_gpio_alert_display_background(AdGpioAlert *sample_filter, Mat &mask)
{
  int i, j;
  uchar *img_ptr, *mask_ptr;
  int n_rows_img = sample_filter->priv->cv_image->rows;
  int n_cols_img = sample_filter->priv->cv_image->cols;

  for (i = 0; i < n_rows_img; ++i)
  {
    img_ptr = sample_filter->priv->cv_image->ptr<uchar>(i);
    mask_ptr = mask.ptr<uchar>(i);

    for (j = 0; j < n_cols_img; ++j)
    {
      img_ptr[j * sample_filter->priv->cv_image->channels()] = mask_ptr[j];
      img_ptr[j * sample_filter->priv->cv_image->channels() + 1] = mask_ptr[j];
      img_ptr[j * sample_filter->priv->cv_image->channels() + 2] = mask_ptr[j];
    }
  }

  return;
}

static void
ad_gpio_alert_initialize_images(AdGpioAlert *sample_filter,
                                     GstVideoFrame *frame, GstMapInfo &info)
{
  if (sample_filter->priv->cv_image == NULL)
  {
    sample_filter->priv->cv_image = new Mat(frame->info.height,
                                            frame->info.width, CV_8UC3, info.data);
  }
  else if ((sample_filter->priv->cv_image->cols != frame->info.width) || (sample_filter->priv->cv_image->rows != frame->info.height))
  {
    delete sample_filter->priv->cv_image;
    sample_filter->priv->cv_image = new Mat(frame->info.height, frame->info.width,
                                            CV_8UC3, info.data);
  }
  else
  {
    sample_filter->priv->cv_image->data = info.data;
  }
}

gboolean
ad_gpio_alert_plugin_init(GstPlugin *plugin)
{
  return gst_element_register(plugin, PLUGIN_NAME, GST_RANK_NONE,
                              AD_TYPE_GPIO_ALERT);
}

#ifndef PACKAGE
#define PACKAGE "SAMPLE"
#endif
#ifndef PACKAGE_VERSION
#define PACKAGE_VERSION "1.0"
#endif
#ifndef GST_PACKAGE_NAME
#define GST_PACKAGE_NAME "Sample Package"
#endif
#ifndef GST_LICENSE
#define GST_LICENSE "LGPL"
#endif
#ifndef GST_PACKAGE_ORIGIN
#define GST_PACKAGE_ORIGIN "https://www.adlinktech.com"
#endif

GST_PLUGIN_DEFINE(
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    adgpioalert,
    "ADLINK GPIO Control Plugin",
    ad_gpio_alert_plugin_init,
    PACKAGE_VERSION,
    GST_LICENSE,
    GST_PACKAGE_NAME,
    GST_PACKAGE_ORIGIN)
